# coding: utf-8

__doc__ = \
"""
*****************************************************************************

                      NAIADES QUESU CSV VALIDATOR
           Outil de contrôle des données au format QUESU CSV 3.1

Ce module définit la classe de validation des fichiers QUESU CSV
"""
"\n*****************************************************************************\n\n"

import io
import os
import shutil
import pathlib
import cchardet
from chardet.universaldetector import UniversalDetector
from quesu_csv_type import *
from quesu_csv_error import QuesuError
from quesu_csv_line import QuesuLine, PC_ANA_PARSE_INFO, PC_CE_PARSE_INFO

class QuesuFile():
    """Classe de validation d'un fichier QUESU"""

    QUESU_PC_ANA_MATCH = "QUESU_CSV_ANA*.csv"
    QUESU_PC_CE_MATCH = "QUESU_CSV_CEP*.csv"
    QUESU_PC_ANA_ERR_MATCH = "QUESU_CSV_ANA_err*.csv"
    QUESU_PC_CE_ERR_MATCH = "QUESU_CSV_CEP_err*.csv"

    QUESU_OUTPUT_DIR_NAME = "quesu_validation"

    def __init__(self, quesu_file_name, quesu_logger):
        """Constructeur

        quesu_file_name : objet pathlib.Path correspondant au fichier QUESU CSV contrôlé
        quesu_logger : objet QuesuErrorLogger chargé d'enregistrer les non conformités détectées
        """
        self.quesu_file_name_src = None # Chemin d'accès au fichier QUESU à traiter
        # Chemin d'accès au fichier QUESU validé = fichier QUESU ré-encodé en UTF8 nettoyé de tout retour à ligne
        # et d'en-tête réduit à la 1ère ligne
        self.quesu_file_name = None
        self.quesu_type = None # Type de données QUESU détecté
        self.quesu_parse_info = None # Paramétrage du parseur pour le type de données QUESU détecté
        self.quesu_logger = None # Objet d'enregistrement des non-conformités

        quesu_file_path = pathlib.Path(quesu_file_name)
        if quesu_file_path.exists() and quesu_file_path.is_file():
            if quesu_file_path.match(__class__.QUESU_PC_ANA_ERR_MATCH):
                self.quesu_type = QUESU_TYPE_PC_ANA_ERR
            elif quesu_file_path.match(__class__.QUESU_PC_CE_ERR_MATCH):
                self.quesu_type = QUESU_TYPE_PC_CE_ERR
            elif quesu_file_path.match(__class__.QUESU_PC_ANA_MATCH):
                self.quesu_type = QUESU_TYPE_PC_ANA
                self.quesu_parse_info = PC_ANA_PARSE_INFO
                self.quesu_empty_content = ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;FLG"
            elif quesu_file_path.match(__class__.QUESU_PC_CE_MATCH):
                self.quesu_type = QUESU_TYPE_PC_CE
                self.quesu_parse_info = PC_CE_PARSE_INFO
                self.quesu_empty_content = ";;;;;;;;;;;;;;;;FLG"
            else:
                self.quesu_type = QUESU_TYPE_INCONNU

            quesu_validation_dir_path = quesu_file_path.parents[0].joinpath(__class__.QUESU_OUTPUT_DIR_NAME)
            quesu_validation_file_path = quesu_validation_dir_path.joinpath(quesu_file_path.name)
            self.quesu_file_name_src = quesu_file_path.as_posix()
            self.quesu_file_name = quesu_validation_file_path.as_posix()
        else:
            self.quesu_type = QUESU_TYPE_INCONNU

        self.quesu_logger = quesu_logger

    def validate(self, quesu_references):
        """Valide le fichier QUESU : nom du fichier, encodage utf8, syntaxe et références Sandre

        quesu_references : références SANDRE obtenues de l'objet QuesuReference
        """
        if self.quesu_type == QUESU_TYPE_INCONNU:
            errors = []
            errors.append(QuesuError(self.quesu_file_name_src, level=QuesuError.ERROR_LEVEL_WARNING,
                                     msg=QuesuError.ERROR_MSG_INVALID_FILE_NAME, line=0))
            self.quesu_logger.log(quesu_file_path=self.quesu_file_name_src, quesu_type=self.quesu_type, errors=errors)
            return {
                'line_count': 0,
                'error_count': self.quesu_logger.file_stats[self.quesu_file_name_src][QuesuError.ERROR_LEVEL_FATAL] +
                               self.quesu_logger.file_stats[self.quesu_file_name_src][QuesuError.ERROR_LEVEL_WARNING] +
                               self.quesu_logger.file_stats[self.quesu_file_name_src][QuesuError.ERROR_LEVEL_INFO],
                'fatal_count': self.quesu_logger.file_stats[self.quesu_file_name_src][QuesuError.ERROR_LEVEL_FATAL],
                'warning_count': self.quesu_logger.file_stats[self.quesu_file_name_src][QuesuError.ERROR_LEVEL_WARNING],
                'info_count': self.quesu_logger.file_stats[self.quesu_file_name_src][QuesuError.ERROR_LEVEL_INFO]}

        # Avant tout traitement, copie du fichier quesu dans le répertoire de travail
        quesu_file = pathlib.Path(self.quesu_file_name_src)
        quesu_validation_file = pathlib.Path(self.quesu_file_name)
        quesu_validation_folder = quesu_validation_file.parents[0]
        if not quesu_validation_folder.exists():
            quesu_validation_folder.mkdir()
        with quesu_file.open(mode='rb') as source:
            with quesu_validation_file.open(mode='wb') as target:
                shutil.copyfileobj(source, target)

        # Validation de l'enveloppe des données : encodage utf8, absence de retour à la ligne et en-tête limité
        # à la 1ère ligne
        self.__validateEncoding(quesu_references=quesu_references)

        # Validation des données : syntaxe et existence des références SANDRE
        return self.__validateData(quesu_references=quesu_references)

    def __validateEncoding(self, quesu_references):
        """Valide l'encodage du fichier QUESU
        Le cas échéant, déclenche les opérations suivantes :
         - réencodage en utf8 pour les fichiers encodés cp1252
         - tronquage de l'en-tête pour le limiter à la 1ère ligne des noms de colonnes
         - suppression des retours à la ligne dans les données
        """
        errors = self.__convertToUTF8()
        if len(errors):
            self.quesu_logger.log(quesu_file_path=self.quesu_file_name, quesu_type=self.quesu_type, errors=errors)
        errors = self.__truncateHeader(quesu_references=quesu_references)
        if len(errors):
            self.quesu_logger.log(quesu_file_path=self.quesu_file_name, quesu_type=self.quesu_type, errors=errors)
        errors = self.__suppressCR()
        if len(errors):
            self.quesu_logger.log(quesu_file_path=self.quesu_file_name, quesu_type=self.quesu_type, errors=errors)

    def __convertToUTF8(self, universal_endline=True):
        """
        Convert any type of file to UTF-8 without BOM
        and using universal endline by default.

        Parameters
        ----------
        universal_endline : boolean (True),
                            by default convert endlines to universal format.
        """
        errors = []

        # Detect encoding
        quesu_file_path = pathlib.Path(self.quesu_file_name)
        with open(self.quesu_file_name, "rb") as file_src:
            if quesu_file_path.stat().st_size > 5e8:
                detector = UniversalDetector()
                for line in file_src:
                    detector.feed(line)
                    if detector.done: break
                detector.close()
                encoding = detector.result['encoding'].upper()
            else:
                content = file_src.read()
                encoding = cchardet.detect(content)['encoding'].upper()

        # Log invalid encoding
        if encoding != "UTF-8" and encoding != 'UTF-8-SIG':
            errors.append(QuesuError(self.quesu_file_name, level=QuesuError.ERROR_LEVEL_WARNING,
                                     msg=QuesuError.ERROR_MSG_INVALID_ENCODING, line=0,
                                     content=self.quesu_empty_content))

        # Convert to utf-8 from file
        with open(self.quesu_file_name, mode="r", encoding=encoding) as file_src:
            with io.open("%s_utf8.csv" % self.quesu_file_name, mode="w", encoding="utf-8") as file_dst:
                i = 0
                for line in file_src:
                    line = line.replace('\r\n', '\n')
                    file_dst.write(line)

        os.replace('%s_utf8.csv' % self.quesu_file_name, self.quesu_file_name)

        return errors

    def __truncateHeader(self, quesu_references):
        """Réduit l'en-tête du fichier QUESU CSV à la 1ère ligne des noms de colonnes

        quesu_references : références SANDRE

        Renvoie une liste d'erreurs rencontrées
        """
        with open(self.quesu_file_name, encoding='utf-8') as file:
            line = file.readline().strip()  # On ne traite pas la 1ère ligne de l'en tête
            line = file.readline().strip()

        quesu_line = QuesuLine(quesu_type=self.quesu_type, quesu_parse_info=self.quesu_parse_info,
                                   quesu_file_name=self.quesu_file_name, quesu_empty_content=self.quesu_empty_content)
        observation, errors = quesu_line.validate(quesu_line=line, quesu_references=quesu_references, quesu_line_num=1)
        if len(errors) > 10:
            # Au-delà de 10 erreurs, on estime que la 2ème ligne du fichier est une ligne d'alias => on génère une
            # nouvelle version du fichier sans cette ligne d'alias
            with io.open(self.quesu_file_name, encoding='utf-8') as source:
                with io.open('%s_trunc.csv' % self.quesu_file_name, mode='w', encoding='utf-8') as target:
                    target.write(source.readline().strip() + "\n")
                    source.readline().strip() # ne tient pas compte de la ligne des alias
                    shutil.copyfileobj(source, target)

            os.replace('%s_trunc.csv' % self.quesu_file_name, self.quesu_file_name)
            errors = []
            errors.append(QuesuError(self.quesu_file_name, level=QuesuError.ERROR_LEVEL_WARNING,
                                     msg=QuesuError.ERROR_MSG_INVALID_HEADER, line=0, content=self.quesu_empty_content))
        else:
            errors = []
        return errors

    def __suppressCR(self):
        """Supprime les retours chariots au sein des données

        Renvoie une liste d'erreurs recontrées
        """
        errors = []
        linebuffer = ""
        with io.open(self.quesu_file_name, encoding='utf8') as src, \
                io.open('%s_cr.csv' % self.quesu_file_name, mode='w', encoding='utf-8') as dst:
            line_count = 0
            for line in src:
                line_count += 1
                linebuffer += line
                if linebuffer.find(QuesuLine.QUESU_EOF) > 0:
                    dst.write(linebuffer)
                    linebuffer = ""
                else:
                    # Suppression du retour chariot
                    linebuffer = linebuffer[0:len(linebuffer) - 2]
                    errors.append(QuesuError(file=self.quesu_file_name, level=QuesuError.ERROR_LEVEL_WARNING,
                                             msg=QuesuError.ERROR_MSG_INVALID_CR, line=line_count,
                                             content=self.quesu_empty_content))

        os.replace('%s_cr.csv' % self.quesu_file_name, self.quesu_file_name)
        return errors

    def __validateHeader(self, header):
        """Valide l'en-tête du fichier QUESU
        
        Renvoie une liste des erreurs rencontrées
        """
        errors = []
        quesu_header = ""
        for element in self.quesu_parse_info:
            quesu_header += "<" + element[0] + ">" + QuesuLine.QUESU_FIELD_SEPARATOR
        quesu_header = quesu_header[0:len(quesu_header) - 1]
        if header != quesu_header:
            errors.append(QuesuError(self.quesu_file_name, level=QuesuError.ERROR_LEVEL_FATAL,
                                     msg=QuesuError.ERROR_MSG_INVALID_HEADER, line=1, content=self.quesu_empty_content))
        return errors

    def __validateData(self, quesu_references):
        """Valide les données QUESU CSV (syntaxe et références SANDRE)
        
        Renvoie un dictionnaire des statistiques de validation comprenant :
        - le nombre de ligne de données contrôlées
        - le nombre total d'erreurs détectées
        - le nombre d'erreurs classées "fatale"
        - le nombre d'erreurs classées "alerte"
        - le nombre d'erreurs classées "information"
        """
        with open(self.quesu_file_name, encoding='utf-8') as quesu_file:

            # Validation de l'en-tête QUESU
            header = quesu_file.readline().strip()
            line_count = 1  # Compteur de ligne de données traitées
            errors = self.__validateHeader(header)
            if len(errors):
                self.quesu_logger.log(quesu_file_path=self.quesu_file_name, quesu_type=self.quesu_type, errors=errors)
                return {
                    'line_count': line_count - 1,
                    'error_count': self.quesu_logger.file_stats[self.quesu_file_name][
                                       QuesuError.ERROR_LEVEL_FATAL] +
                                   self.quesu_logger.file_stats[self.quesu_file_name][
                                       QuesuError.ERROR_LEVEL_WARNING] +
                                   self.quesu_logger.file_stats[self.quesu_file_name][
                                       QuesuError.ERROR_LEVEL_INFO],
                    'fatal_count': self.quesu_logger.file_stats[self.quesu_file_name][QuesuError.ERROR_LEVEL_FATAL],
                    'warning_count': self.quesu_logger.file_stats[self.quesu_file_name][
                        QuesuError.ERROR_LEVEL_WARNING],
                    'info_count': self.quesu_logger.file_stats[self.quesu_file_name][QuesuError.ERROR_LEVEL_INFO]}

            # Validation des enregistrements QUESU
            quesu_line = QuesuLine(quesu_type=self.quesu_type, quesu_parse_info=self.quesu_parse_info,
                                   quesu_file_name=self.quesu_file_name, quesu_empty_content=self.quesu_empty_content)
            for line in quesu_file:
                line_count += 1
                observation, errors = quesu_line.validate(quesu_line=line.strip(), quesu_references=quesu_references,
                                                          quesu_line_num=line_count)
                if len(errors):
                    self.quesu_logger.log(quesu_file_path=self.quesu_file_name, quesu_type=self.quesu_type,
                                          errors=errors, force_write=False)

        # Force l'écriture des non-conformités en mémoire
        self.quesu_logger.log(quesu_file_path=self.quesu_file_name, quesu_type=self.quesu_type, errors=[],
                              force_write=True)

        if self.quesu_file_name in self.quesu_logger.file_stats.keys():
            return {
                'line_count': line_count - 1,
                'error_count': self.quesu_logger.file_stats[self.quesu_file_name][QuesuError.ERROR_LEVEL_FATAL] +
                               self.quesu_logger.file_stats[self.quesu_file_name][
                                   QuesuError.ERROR_LEVEL_WARNING] + self.quesu_logger.file_stats[self.quesu_file_name][
                                   QuesuError.ERROR_LEVEL_INFO],
                'fatal_count': self.quesu_logger.file_stats[self.quesu_file_name][QuesuError.ERROR_LEVEL_FATAL],
                'warning_count': self.quesu_logger.file_stats[self.quesu_file_name][QuesuError.ERROR_LEVEL_WARNING],
                'info_count': self.quesu_logger.file_stats[self.quesu_file_name][QuesuError.ERROR_LEVEL_INFO]}
        else:
            return {
                'line_count': line_count - 1,
                'error_count': 0,
                'fatal_count': 0,
                'warning_count': 0,
                'info_count': 0}
