# coding: utf-8

__doc__ = \
"""
*****************************************************************************

                      NAIADES QUESU CSV VALIDATOR
           Outil de contrôle des données au format QUESU CSV 3.1

Ce module regroupe les objets de chargement des référentiels du Sandre

"\n*****************************************************************************\n\n"
"""

import io
import urllib.request
import gzip
import shutil
from pathlib import Path
from datetime import datetime
from lxml import objectify
import cherrypy

sandre_reference_ws = {
    'STQ':"https://api.sandre.eaufrance.fr/referentiels/v1/stq.xml?outputSchema=SANDREv3.1&compress=true",
    'DC': "https://api.sandre.eaufrance.fr/referentiels/v1/dc.xml?outputSchema=SANDREv4&compress=true",
    'ProjectPrel': "https://api.sandre.eaufrance.fr/referentiels/v1/nsa.xml?outputSchema=SANDREv3.1&compress=true&filter=%3CFilter%3E%3CIS%3E%3CField%3ECdReferentiel%3C%2FField%3E%3CValue%3E22%3C%2FValue%3E%3C%2FIS%3E%3C%2FFilter%3E",
    'ZoneVerticaleProspectee': "https://api.sandre.eaufrance.fr/referentiels/v1/nsa.xml?outputSchema=SANDREv3.1&compress=true&filter=%3CFilter%3E%3CIS%3E%3CField%3ECdReferentiel%3C%2FField%3E%3CValue%3E430%3C%2FValue%3E%3C%2FIS%3E%3C%2FFilter%3E",
    'DifficultePrel': "https://api.sandre.eaufrance.fr/referentiels/v1/nsa.xml?outputSchema=SANDREv3.1&compress=true&filter=%3CFilter%3E%3CIS%3E%3CField%3ECdReferentiel%3C%2FField%3E%3CValue%3E67%3C%2FValue%3E%3C%2FIS%3E%3C%2FFilter%3E",
    'AccredPrel': "https://api.sandre.eaufrance.fr/referentiels/v1/nsa.xml?outputSchema=SANDREv3.1&compress=true&filter=%3CFilter%3E%3CIS%3E%3CField%3ECdReferentiel%3C%2FField%3E%3CValue%3E333%3C%2FValue%3E%3C%2FIS%3E%3C%2FFilter%3E",
    'FinalitePrel': "https://api.sandre.eaufrance.fr/referentiels/v1/nsa.xml?outputSchema=SANDREv3.1&compress=true&filter=%3CFilter%3E%3CIS%3E%3CField%3ECdReferentiel%3C%2FField%3E%3CValue%3E645%3C%2FValue%3E%3C%2FIS%3E%3C%2FFilter%3E",
    'SUP': "https://api.sandre.eaufrance.fr/referentiels/v1/sup.xml?outputSchema=SANDREv2&compress=true",
    'MET': "https://api.sandre.eaufrance.fr/referentiels/v1/met.xml?outputSchema=SANDREv2&compress=true",
    'INT': "https://api.sandre.eaufrance.fr/referentiels/v1/int.xml?outputSchema=SANDREv2&compress=true",
    'PAR': "https://api.sandre.eaufrance.fr/referentiels/v1/par.xml?outputSchema=SANDREv2&compress=true",
    'FAN': "https://api.sandre.eaufrance.fr/referentiels/v1/fan.xml?outputSchema=SANDREv2&compress=true",
    'URF': "https://api.sandre.eaufrance.fr/referentiels/v1/urf.xml?outputSchema=SANDREv2&compress=true",
    'Rq': "https://api.sandre.eaufrance.fr/referentiels/v1/nsa.xml?outputSchema=SANDREv3.1&compress=true&filter=%3CFilter%3E%3CIS%3E%3CField%3ECdReferentiel%3C%2FField%3E%3CValue%3E155%3C%2FValue%3E%3C%2FIS%3E%3C%2FFilter%3E",
    'InsituAna': "https://api.sandre.eaufrance.fr/referentiels/v1/nsa.xml?outputSchema=SANDREv3.1&compress=true&filter=%3CFilter%3E%3CIS%3E%3CField%3ECdReferentiel%3C%2FField%3E%3CValue%3E156%3C%2FValue%3E%3C%2FIS%3E%3C%2FFilter%3E",
    'DifficulteAna': "https://api.sandre.eaufrance.fr/referentiels/v1/nsa.xml?outputSchema=SANDREv3.1&compress=true&filter=%3CFilter%3E%3CIS%3E%3CField%3ECdReferentiel%3C%2FField%3E%3CValue%3E43%3C%2FValue%3E%3C%2FIS%3E%3C%2FFilter%3E",
    'Qual': "https://api.sandre.eaufrance.fr/referentiels/v1/nsa.xml?outputSchema=SANDREv3.1&compress=true&filter=%3CFilter%3E%3CIS%3E%3CField%3ECdReferentiel%3C%2FField%3E%3CValue%3E414%3C%2FValue%3E%3C%2FIS%3E%3C%2FFilter%3E",
    'Statut': "https://api.sandre.eaufrance.fr/referentiels/v1/nsa.xml?outputSchema=SANDREv3.1&compress=true&filter=%3CFilter%3E%3CIS%3E%3CField%3ECdReferentiel%3C%2FField%3E%3CValue%3E446%3C%2FValue%3E%3C%2FIS%3E%3C%2FFilter%3E",
    'AccreAna': "https://api.sandre.eaufrance.fr/referentiels/v1/nsa.xml?outputSchema=SANDREv3.1&compress=true&filter=%3CFilter%3E%3CIS%3E%3CField%3ECdReferentiel%3C%2FField%3E%3CValue%3E299%3C%2FValue%3E%3C%2FIS%3E%3C%2FFilter%3E"
}

quesu_reference_fields = {
    'STQ':('CdStationMesureEauxSurface', 'CdPointEauxSurf'),
    'DC':('CodeSandreRdd',),
    'ProjectPrel':('ProjectPrel',),
    'ZoneVerticaleProspectee':('ZoneVerticaleProspectee',),
    'DifficultePrel':('DifficultePrel',),
    'AccredPrel':('AccredPrel',),
    'FinalitePrel':('FinalitePrel',),
    'SUP':('CdSupport',),
    'MET':('CdMethodePrel', 'CdMethFractionnement', 'CdMethAna', 'CdMethExtraction', 'CdMethode'),
    'INT':('ProducteurPrel', 'Preleveur', 'ProducteurAna', 'Laboratoire', 'Producteur'),
    'PAR':('CdParametre', 'RsAna', 'RsParEnv'),
    'FAN':('CdFractionAnalysee',),
    'URF':('CdUniteMesure',),
    'Rq':('RqAna', 'RqParEnv',),
    'InsituAna':('InsituAna',),
    'DifficulteAna':('DifficulteAna',),
    'Qual':('QualAna', 'QualParEnv',),
    'Statut':('StatutAna', 'StatutParEn'),
    'AccreAna':('AccreAna',)
}


class QuesuReference():
    QUESU_REFERENCE_DIR_NAME = "quesu_reference"
    QUESU_REFERENCE_COUNT = 34

    def __init__(self, quesu_path):
        """Initalise l'objet en chargeant l'ensemble des référentiels dans une attribut d'objet
        L'objet Path "quesu_path" passé en paramètre fournit le chemin d'accès au répertoire utilisé pour le stockage des fichiers gzip et xml téléchargés
        """
        self.quesu_ref_path = None
        
        if quesu_path.exists() and quesu_path.is_dir() :
            # Création du répertoire de stockage des fichiers gzip et xml des référentiels
            quesu_ref_path = quesu_path.joinpath(__class__.QUESU_REFERENCE_DIR_NAME)
            if not quesu_ref_path.exists():
                quesu_ref_path.mkdir()
            self.quesu_ref_path = quesu_ref_path
            self.references = self.__getReferences()

    def getReferences(self):
        """Renvoie les référentiels Sandre"""
        if self.quesu_ref_path:
            return self.references
        return None

    def __getReferences(self):
        """Renvoie l'ensemble des référentiels SANDRE"""
        #cherrypy.log("{0} : => Récupération des références SANDRE...".format(datetime.now().strftime("%d/%m/%y %H:%M:%S")))
        quesu_refs = {}
        for sandre_ref_id, sandre_ref_url in sandre_reference_ws.items():
            cherrypy.log("{0} : => Service web {1}:{2}".format(datetime.now().strftime("%d/%m/%y %H:%M:%S"), sandre_ref_id, sandre_ref_url))
            sandre_tree = self.__getSandreTree(ref_id=sandre_ref_id, ref_url=sandre_ref_url)
            if sandre_tree is None:
                return None

            for quesu_reference_field in quesu_reference_fields[sandre_ref_id]:
                #cherrypy.log("{0} : => Extraction des références du champ {1}".format(
                #    datetime.now().strftime("%d/%m/%y %H:%M:%S"), quesu_reference_field))
                quesu_refs.update(self.__getQuesuReferences(quesu_reference_field=quesu_reference_field,
                                                             sandre_tree=sandre_tree))
        return quesu_refs

    def __getQuesuReferences(self, quesu_reference_field, sandre_tree):
        """Renvoie les références associées à un champ de données du format QUESU CSV

        quesu_reference_id : identifiant du champ QUESU CSV
        sandre_tree : arborescence d'objets ObjectifiedElement du référentiel SANDRE interrogé

        Renvoie un tuple des références dans un dictionnaire
        """
        refs = {}
        root = sandre_tree.getroot()
        if quesu_reference_field == "CdStationMesureEauxSurface":
            #refs[quesu_reference_field] = (station.CdStationMesureEauxSurface.text for station in root.Referentiel.StationMesureEauxSurface)
            cdstations = []
            for station in root.Referentiel.StationMesureEauxSurface:
                cdstations.append(station.CdStationMesureEauxSurface.text)
            refs[quesu_reference_field] = tuple(cdstations)
        elif quesu_reference_field == "CdPointEauxSurf":
            for station in root.Referentiel.StationMesureEauxSurface:
                cdstation = station.CdStationMesureEauxSurface.text
                if hasattr(station, "PointPrelEauxSurf"):
                    #refs[quesu_reference_field + "_" + cdstation] = (pointprel.CdPointEauxSurf.text for pointprel in station.PointPrelEauxSurf)
                    cdpointprels = []
                    for pointprel in station.PointPrelEauxSurf:
                        cdpointprels.append(pointprel.CdPointEauxSurf.text)
                    refs[quesu_reference_field + "_" + cdstation] = tuple(cdpointprels)
                else:
                    refs[quesu_reference_field + "_" + cdstation] = ()
        elif quesu_reference_field == "CodeSandreRdd":
            #refs[quesu_reference_field] = (reseau.CodeSandreRdd.text for reseau in root.Referentiel.DispositifCollecte)
            cdsandrerdds = []
            for reseau in root.Referentiel.DispositifCollecte:
                cdsandrerdds.append(reseau.CodeSandreRdd.text)
            refs[quesu_reference_field] = tuple(cdsandrerdds)
        elif quesu_reference_field in (
                "ProjectPrel", "ZoneVerticaleProspectee", "DifficultePrel", "AccredPrel", "FinalitePrel", "RqAna",
                "InsituAna",
                "DifficulteAna", "QualAna", "StatutAna", "AccreAna"):
            #refs[quesu_reference_field] = (ref.CdElement.text for ref in root.Referentiel.Element)
            cdelements = []
            for ref in root.Referentiel.Element:
                cdelements.append(ref.CdElement.text)
            refs[quesu_reference_field] = tuple(cdelements)
        elif quesu_reference_field == "CdSupport":
            #refs[quesu_reference_field] = (support.CdSupport.text for support in root.Referentiel.Support)
            cdsupports =[]
            for support in root.Referentiel.Support:
                cdsupports.append(support.CdSupport.text)
            refs[quesu_reference_field] = tuple(cdsupports)
        elif quesu_reference_field in ("CdMethodePrel", "CdMethFractionnement", "CdMethAna", "CdMethExtraction", "CdMethode"):
            #refs[quesu_reference_field] = (methode.CdMethode.text for methode in root.Referentiel.Methode)
            cdmethodes = []
            for methode in root.Referentiel.Methode:
                cdmethodes.append(methode.CdMethode.text)
            refs[quesu_reference_field] = tuple(cdmethodes)
        elif quesu_reference_field in ("ProducteurPrel", "Preleveur", "ProducteurAna", "Laboratoire", "Producteur"):
            #refs[quesu_reference_field] = (intervenant.CdIntervenant.text for intervenant in root.Referentiel.Intervenant)
            cdintervenants = []
            for intervenant in root.Referentiel.Intervenant:
                cdintervenants.append(intervenant.CdIntervenant.text)
            refs[quesu_reference_field] = tuple(cdintervenants)
        elif quesu_reference_field in ("CdParametre"):
            cdparametres = []
            for parametre in root.Referentiel.Parametre:
                cdparametres.append(parametre.CdParametre.text)
            refs[quesu_reference_field] = tuple(cdparametres)
        elif quesu_reference_field in ("RsAna", "RsParEnv"):
            nsmap = {k: v for k, v in root.nsmap.items() if k is None}
            path = ".//{" + nsmap[None] + "}CdValeursPossiblesParametre"
            for parametre in root.Referentiel.Parametre:
                cdparametre = parametre.CdParametre.text
                # Recherche de la balise <CdValeursPossiblesParametre> n'importe où sous la balise du parametre courant
                # IMPORTANT : il faut préciser l'espace de noms principal dans le paramètre "path" de la méthode findall,
                # autrement la recherche est infructueuse
                cdvaleurspossibles = [value.text for value in parametre.findall(path=path)]
                if len(cdvaleurspossibles):
                    refs[quesu_reference_field + "_" + cdparametre] = tuple(cdvaleurspossibles)
        elif quesu_reference_field in ("CdFractionAnalysee"):
            #refs[quesu_reference_field] = (fraction.CdFractionAnalysee.text for fraction in root.Referentiel.FractionAnalysee)
            cdfractionanalysees = []
            for fraction in root.Referentiel.FractionAnalysee:
                cdfractionanalysees.append(fraction.CdFractionAnalysee.text)
            refs[quesu_reference_field] = tuple(cdfractionanalysees)
        elif quesu_reference_field in ("CdUniteMesure"):
            #refs[quesu_reference_field] = (unite.CdUniteReference.text for unite in root.Referentiel.UniteReference)
            cdunites = []
            for unite in root.Referentiel.UniteReference:
                cdunites.append(unite.CdUniteReference.text)
            refs[quesu_reference_field] = tuple(cdunites)

        return refs

    def __getSandreTree(self, ref_id, ref_url):
        """Renvoie l'arborescence d'objets d'un référentiel SANDRE

        ref_id : identifiant du référentiel
        ref_url : url du service web de diffusion du référentiel

        Renvoie une arborescence d'objets ObjectifiedElement
        """
        # Initialisation des chemins d'accès aux fichiers zip et xml du référentiel téléchargé
        try:
            zip_file_name = self.quesu_ref_path.joinpath(ref_id + ".xml.gz").as_posix()
            xml_file_name = self.quesu_ref_path.joinpath(ref_id + ".xml").as_posix()
            # Initialisation du dictionnaire résultat
            refs = {}

            # Optimisation des délais de traitement : l'interrogation des services web n'est déclenchée que si les fichiers
            # xml sont inexistants ou qu'ils datent de la veille
            upload_needed = False
            xml_file_path = Path(xml_file_name)
            if xml_file_path.exists():
                file_dt = datetime.fromtimestamp(xml_file_path.stat().st_mtime).date()
                if file_dt != datetime.now().date():
                    upload_needed = True
            else:
                upload_needed = True

            if upload_needed:
                # Récupération des données du référentiel via interrogation du service web SANDRE
                cherrypy.log("{0} : => Interrogation du service web {1}".format(datetime.now().strftime("%d/%m/%y %H:%M:%S"), ref_url))
                zip_data = None
                try:
                    with urllib.request.urlopen(ref_url) as response:
                        zip_data = response.read()
                except Exception as ex:
                    cherrypy.log("{0} : => Erreur d'interrogation {1}".format(
                        datetime.now().strftime("%d/%m/%y %H:%M:%S"), ex))

                if zip_data:
                    cherrypy.log("{0} : => Sauvegarde des données compressées dans {1}".format(
                        datetime.now().strftime("%d/%m/%y %H:%M:%S"), zip_file_name))
                    with io.open(zip_file_name, "wb") as f:
                        f.write(zip_data)

                    cherrypy.log(
                        "{0} : => Décompression des données dans {1}".format(datetime.now().strftime("%d/%m/%y %H:%M:%S"),
                                                                             xml_file_name))
                    with io.open(xml_file_name, 'wb') as xml_file:
                        with gzip.open(zip_file_name, 'rb') as zip_file:
                            shutil.copyfileobj(zip_file, xml_file)
                else:
                    cherrypy.log(
                        "{0} : => Exploitation d'un référentiel antérieur".format(datetime.now().strftime("%d/%m/%y %H:%M:%S")))

            xml_file_path = Path(xml_file_name)
            if not xml_file_path.exists():
                cherrypy.log("{0} : => Fichier {1} introuvale".format(datetime.now().strftime("%d/%m/%y %H:%M:%S"), xml_file_name))
                return None

            xml_tree = objectify.parse(xml_file_name)
            return xml_tree
        except:
            return None


if __name__ == '__main__':
    quesu_path = Path("C:/_data_naiades/test")
    quesu_refs = QuesuReference(quesu_path).getReferences()
    for key, values in quesu_refs.items():
        for value in values:
            cherrypy.log("%s : %s" % (key, value))